# Python multivariate analysis (Support Vector Classification) script
# written by Simon O'Callaghan NICTA
# 14 May 2015

import pandas as pd
import numpy as np
import utils
from matplotlib import pyplot as pl
from computers import gp
from mpl_toolkits.mplot3d import Axes3D
from sklearn.linear_model import LogisticRegression

from revrand.utils.datasets import fetch_gpml_usps_resampled_data
from revrand.validation import loglosscat, errrate
from revrand import classification, basis_functions, glm, likelihoods
import logging
logging.basicConfig(level=logging.DEBUG)


def loadData(path):
    data = pd.read_csv(path)
    for i in range(len(data.columns)):
      data  = data[np.isfinite(data[data.columns[i]])]  # Remove any nans
    # Set up the features (X) and the targets or labels (y)
    X = data[data.columns[3:]].values
    y = data.lithology.values
    lonlat = data[data.columns[:2]].values
    return X,y, lonlat


def preprocessData(X):
    # take log of features with a high dynamic range
    X[:, 4] = np.log(X[:, 4]+1e-20)
    X[:, 6] = np.log(X[:, 6]+1e-20)
    return X

def normaliseData(X):
    means = np.mean(X, axis=0)
    std_devs = np.std(X, axis=0)
    X = (X - means)/std_devs
    return X, means, std_devs

def extractUncorFeatures(X):
    # keepdims = [0,6,4,5,7, -1]
    keepdims = [0,2,4,5,7, -1]
    X = X[:,keepdims]
    allFeaturesList = ['Depth', 'nitrate', 'Oxygen', 'Phosphate', 'log(Productivity)', 'Salinity', 'log(Silicate)','Temperature']

    return X, [allFeaturesList[id] for id in keepdims[:5]]

def removeHemisphere(X, y_map, lonlat, hemToRemove = 'north'):
    if hemToRemove == 'north':
        keeppoints_id= np.where(lonlat[:,1]<0)[0]
        locStr = '_north'
    elif hemToRemove == 'south':
        keeppoints_id= np.where(lonlat[:,1]>0)[0]
        locStr = '_south'
    else:
        raise('need to specify an hemisphere to remove')

    return X[keeppoints_id,:], y_map[keeppoints_id], locStr

def sat_outliers(data, m=3):
     data[(data - np.mean(data)) > m * np.std(data)] = np.mean(data) + \
                                                       m * np.std(data)
     data[(data - np.mean(data)) < - m * np.std(data)] = np.mean(data) - \
                                                         m * np.std(data)
     return data

def kerneldef2D(h, k):
    return h(1e-3, 1e5, 1) * k('matern3on2', np.asarray([h(5e-1, 1e1, 1)
                                      for _ in range(2)]))

def kerneldef3D(h, k):
    return h(1e-3, 1e5, 1) * k('matern3on2', np.asarray([h(5e-1, 1e1, 1)
                                      for _ in range(3)]))

def plotscatter2D(X,y,featureID, featureLabels, locStr):
    for i in range(2):

        featureOfInt = sorted([featureID[0], featureID[i+1]])

        fig = pl.figure()
        ax = fig.add_subplot(111)
        negInd = np.where(y == 0.)[0]
        ax.scatter(X[negInd,featureOfInt[0]], X[negInd,featureOfInt[1]], c='b', alpha=0.15, label='Negative occurrences')
        posyInd = np.where(y == 1.)[0]
        ax.scatter(X[posyInd,featureOfInt[0]], X[posyInd,featureOfInt[1]], c='r',label='Positive occurrences')
        ax.legend(loc=1)
        ax.set_xlim((np.min(X[:,featureOfInt[0]]),np.max(X[:, featureOfInt[0]])))
        if featureID[0] == 0: # Set scale for bathymetry requested by Adriana
            ax.set_xlim((0,7500))
        ax.set_ylim((np.min(X[:,featureOfInt[1]]),np.max(X[:, featureOfInt[1]])))
        pl.xlabel(str(featureLabels[featureOfInt[0]]))
        pl.ylabel(str(featureLabels[featureOfInt[1]]))
        pl.savefig('scatter_Litho' + str(lithoLabel) + locStr + '_' + str(featureLabels[featureOfInt[0]])
                   + '_' + str(featureLabels[featureOfInt[1]]) + '.pdf')

def plotscatter3D(X, y, featureID, featureLabels, locStr):
    # sort features in terms of their index (requested by adriana)
    featureIDsortID = np.argsort(featureID)
    featureIDsort = [featureID[ID] for ID in featureIDsortID]

    fig = pl.figure()
    ax = fig.add_subplot(111, projection='3d')
    negInd = np.where(y == 0.)[0]
    ax.scatter(X[negInd,featureIDsort[0]], X[negInd,featureIDsort[1]], X[negInd,featureIDsort[2]], c='b', alpha=0.15, label='Negative occurrences')
    posyInd = np.where(y == 1.)[0]
    ax.scatter(X[posyInd,featureIDsort[0]], X[posyInd,featureIDsort[1]], X[posyInd,featureIDsort[2]], c='r',label='Positive occurrences')
    # ax.legend(loc=4)
    ax.legend()
    ax.set_xlim((np.min(X[:,featureIDsort[0]]),np.max(X[:,featureIDsort[0]])))
    ax.set_ylim((np.min(X[:,featureIDsort[1]]),np.max(X[:,featureIDsort[1]])))
    pl.xlabel(str(featureLabels[featureIDsort[0]]))
    pl.ylabel(str(featureLabels[featureIDsort[1]]))
    ax.set_zlabel(str(featureLabels[featureIDsort[2]]))
    pl.savefig('scatter3d'+str(lithoLabel) + locStr + '_'+ str(featureLabels[featureIDsort[0]]) +'_'+ str(featureLabels[featureIDsort[1]]) + '_'+
              str(featureLabels[featureIDsort[2]]) + '.pdf')

def plotprobrevrand(X_all,Y,featureID,featureLabels,locStr, ntraining, Xorigscale):
    # A la Carte classifier setting
    nbases = 600
    lenscale = 1
    reg = 1e3
    doSGD = True
    method = 'GLM'
    passes = 300
    K = 5

    for i in range(2):
        featureOfInt = sorted([featureID[0], featureID[i+1]])
        xeva, yeva = np.meshgrid(np.linspace(np.min(X_all[:,featureOfInt[0]]),np.max(X_all[:,featureOfInt[0]])*1.5,test_resol_2D),
                                 np.linspace(np.min(X_all[:,featureOfInt[1]]),np.max(X_all[:,featureOfInt[1]]),np.floor(test_resol_2D)))
        Xs = np.vstack((xeva.flatten(),yeva.flatten())).T
        X = np.vstack((X_all[:,featureOfInt[0]],X_all[:,featureOfInt[1]])).T

        Phi = basis_functions.RandomRBF(nbases, X.shape[1])
        if method == 'SGD':
            weights, labels = classification.learn_sgd(X, Y, Phi, (lenscale,),
                                                       regulariser=reg,
                                                       passes=passes)
        elif method == 'MAP':
            weights, labels = classification.learn_map(X, Y, Phi, (lenscale,),
                                                       regulariser=reg)
        elif method == 'GLM':
            llhood = likelihoods.Bernoulli()
            lparams = []
            params = glm.learn(X, Y, llhood, lparams, Phi, [lenscale], reg=reg,
                               use_sgd=doSGD, maxit=passes, postcomp=K)
        else:
            raise ValueError("Invalid method chosen!")


        lreg = LogisticRegression(penalty='l2', class_weight='auto', C=reg)
        lreg.fit(Phi(X, lenscale), Y)


        # Predict
        # import ipdb; ipdb.set_trace()
        if method == 'GLM':
            pys_l, Vpy, Epn, Epx = glm.predict_meanvar(Xs, llhood, Phi, *params)
            pys_l = np.vstack((1 - pys_l, pys_l)).T
        else:
            pys_l = classification.predict(Xs, weights, Phi, (lenscale,))

        aspect = (np.max(Xorigscale[:,featureOfInt[0]])-np.min(Xorigscale[:,featureOfInt[0]]))/(np.max(Xorigscale[:,featureOfInt[1]]) - np.min(Xorigscale[:,featureOfInt[1]]))
        pl.figure()
        # import ipdb; ipdb.set_trace()
        pl.imshow(np.reshape(pys_l[:,1],xeva.shape), origin='lower', interpolation='nearest',
                  extent=[np.min(Xorigscale[:,featureOfInt[0]]),np.max(Xorigscale[:,featureOfInt[0]]),np.min(Xorigscale[:,featureOfInt[1]]),np.max(Xorigscale[:,featureOfInt[1]])]);
        pl.xlabel(str(featureLabels[featureOfInt[0]]))
        pl.ylabel(str(featureLabels[featureOfInt[1]]))
        pl.colorbar()
        ax = pl.gca()
        if featureID[0] == 0: # Set scale for bathymetry requested by Adriana
            ax.set_xlim((0,7500))
        ax.tick_params(axis='x', top='off')
        ax.tick_params(axis='y', right='off')
        ax.get_yaxis().set_tick_params(direction='out')
        ax.get_xaxis().set_tick_params(direction='out')
        ax.set_aspect(aspect)
        pl.savefig('probLitho'+str(lithoLabel)+locStr+'_'+str(featureLabels[featureOfInt[0]])+'_'+str(featureLabels[featureOfInt[1]])+'.pdf')




def plotprob2D(X,y,featureID,featureLabels,locStr, ntraining, Xorigscale):
    # Generate query points
    for i in range(2):
        featureOfInt = sorted([featureID[0], featureID[i+1]])
        xeva, yeva = np.meshgrid(np.linspace(np.min(X[:,featureOfInt[0]]),np.max(X[:,featureOfInt[0]])*1.5,test_resol_2D),
                                 np.linspace(np.min(X[:,featureOfInt[1]]),np.max(X[:,featureOfInt[1]]),np.floor(test_resol_2D)))

        X_Test_Selected = np.vstack((xeva.flatten(),yeva.flatten())).T

        # Subsample the data to generate a test set
        randInd = np.random.choice(X.shape[0], X.shape[0])
        X_train = X[randInd[:ntraining],:]
        y_train = y[randInd[:ntraining]]

        # Intialise constants
        WALLTIME = 1000.0
        N_CORES = 1

        # Modeling options
        WHITEFN = 'pca'
        APPROX_METHOD = 'laplace'       # 'laplace' or 'pls'
        MULTI_METHOD = 'OVA'            # 'AVA' or 'OVA'
        FUSE_METHOD = 'EXCLUSION'       # 'MODE' or 'EXCLUSION'
        RESPONSE_NAME = 'probit'        # 'probit' or 'logistic'

        optimiser_config = gp.OptConfig()
        optimiser_config.sigma = gp.auto_range(kerneldef2D)
        optimiser_config.walltime = WALLTIME
        responsefunction = gp.classifier.responses.get(RESPONSE_NAME)

        # import ipdb; ipdb.set_trace()
        X_train2D = np.vstack((X_train[:,featureOfInt[0]],X_train[:,featureOfInt[1]])).T
        learned_classifier = gp.classifier.learn(X_train2D, y_train, kerneldef2D,
            responsefunction, optimiser_config,
            multimethod=MULTI_METHOD, approxmethod=APPROX_METHOD,
            train=True, ftol=1e-6, processes=N_CORES)

        print_function = gp.describer(kerneldef2D)
        gp.classifier.utils.print_learned_kernels(print_function, learned_classifier, np.unique(y_train))

        # Print the matrix of learned classifier hyperparameters
        gp.classifier.utils.print_hyperparam_matrix(learned_classifier)

        # predictor_plt = gp.classifier.query(learned_classifier, X)


        predictor_plt = gp.classifier.query(learned_classifier, X_Test_Selected)
        exp_plt = gp.classifier.expectance(learned_classifier, predictor_plt)
        var_plt = gp.classifier.variance(learned_classifier, predictor_plt)
        yq_lde_plt = gp.classifier.linearised_model_differential_entropy(exp_plt, var_plt,
            learned_classifier)
        yq_sd_plt = gp.classifier.equivalent_standard_deviation(yq_lde_plt)
        yq_prob_plt = gp.classifier.predict_from_latent(exp_plt, var_plt,
            learned_classifier, fusemethod=FUSE_METHOD)

        # import ipdb; ipdb.set_trace()
        aspect = (np.max(Xorigscale[:,featureOfInt[0]])-np.min(Xorigscale[:,featureOfInt[0]]))/(np.max(Xorigscale[:,featureOfInt[1]]) - np.min(Xorigscale[:,featureOfInt[1]]))
        pl.figure()
        pl.imshow(np.reshape(yq_prob_plt,xeva.shape), origin='lower', interpolation='nearest',
                  extent=[np.min(Xorigscale[:,featureOfInt[0]]),np.max(Xorigscale[:,featureOfInt[0]]),np.min(Xorigscale[:,featureOfInt[1]]),np.max(Xorigscale[:,featureOfInt[1]])]);
        pl.xlabel(str(featureLabels[featureOfInt[0]]))
        pl.ylabel(str(featureLabels[featureOfInt[1]]))
        pl.colorbar()
        ax = pl.gca()
        if featureID[0] == 0: # Set scale for bathymetry requested by Adriana
            ax.set_xlim((0,7500))
        ax.tick_params(axis='x', top='off')
        ax.tick_params(axis='y', right='off')
        ax.get_yaxis().set_tick_params(direction='out')
        ax.get_xaxis().set_tick_params(direction='out')
        ax.set_aspect(aspect)
        pl.savefig('probLitho'+str(lithoLabel)+locStr+'_'+str(featureLabels[featureOfInt[0]])+'_'+str(featureLabels[featureOfInt[1]])+'.pdf')

        # import IPython; IPython.embed()
        pl.figure()
        pl.imshow(np.reshape(var_plt,xeva.shape), origin='lower', interpolation='nearest',
                  extent=[np.min(Xorigscale[:,featureOfInt[0]]),np.max(Xorigscale[:,featureOfInt[0]]),np.min(Xorigscale[:,featureOfInt[1]]),np.max(Xorigscale[:,featureOfInt[1]])]);
        pl.xlabel(str(featureLabels[featureOfInt[0]]))
        pl.ylabel(str(featureLabels[featureOfInt[1]]))
        pl.colorbar()
        ax = pl.gca()
        if featureID[0] == 0: # Set scale for bathymetry requested by Adriana
            ax.set_xlim((0,7500))
        ax.tick_params(axis='x', top='off')
        ax.tick_params(axis='y', right='off')
        ax.get_yaxis().set_tick_params(direction='out')
        ax.get_xaxis().set_tick_params(direction='out')
        ax.set_aspect(aspect)
        pl.savefig('varLitho'+str(lithoLabel)+locStr+'_'+str(featureLabels[featureOfInt[0]])+'_'+str(featureLabels[featureOfInt[1]])+'.pdf')

        # import IPython; IPython.embed()
        from sklearn.neighbors import NearestNeighbors
        nbrs = NearestNeighbors(n_neighbors=1, algorithm='ball_tree').fit(X_train2D)
        distances, indices = nbrs.kneighbors(X_Test_Selected)

        # make colormap and set alpha channel inversely proportional to distance from training point
        max_dist_thresh = 1
        jet = pl.get_cmap('jet')
        prob_rgb = jet(yq_prob_plt)
        max_dist_sat = distances.copy()
        max_dist_sat[np.where(distances>max_dist_thresh)] = max_dist_thresh
        max_dist_sat_norm = max_dist_sat/np.max(max_dist_sat)
        prob_rgb[:,3] = 1-max_dist_sat_norm[:,0]

        pl.figure()
        pl.imshow(np.reshape(prob_rgb,(xeva.shape[0],xeva.shape[1], 4)), origin='lower', interpolation='nearest',
                  extent=[np.min(Xorigscale[:,featureOfInt[0]]),np.max(Xorigscale[:,featureOfInt[0]]),np.min(Xorigscale[:,featureOfInt[1]]),np.max(Xorigscale[:,featureOfInt[1]])]);
        pl.xlabel(str(featureLabels[featureOfInt[0]]))
        pl.ylabel(str(featureLabels[featureOfInt[1]]))
        pl.colorbar()
        ax = pl.gca()
        if featureID[0] == 0: # Set scale for bathymetry requested by Adriana
            ax.set_xlim((0,7500))

        ax.tick_params(axis='x', top='off')
        ax.tick_params(axis='y', right='off')
        ax.get_yaxis().set_tick_params(direction='out')
        ax.get_xaxis().set_tick_params(direction='out')
        ax.set_aspect(aspect)
        pl.savefig('maskedLitho'+str(lithoLabel)+locStr+'_'+str(featureLabels[featureOfInt[0]])+'_'+str(featureLabels[featureOfInt[1]])+'.pdf')


path = 'seafloor_lith_data_all_features.csv'
lithoLabel = 1
featureID = [1, 5, 4] # List features in order of priority. Base 1 index
nRandomFeatures = 1
removeLabels = [9,10,11,12] # List of lithologies which Adriana said to remove
removeHemisphereFlag = False
hemToRemove = 'south' # choose 'north or south'
balanceDataFlag = True
ntraining = 1500
test_resol_2D = 150

# Load Data
X,y, lonlat = loadData(path)
X[:,0] = np.abs(X[:,0]) # Turn bathymetry into a depth

# Remove a hemisphere from the data
if removeHemisphereFlag:
    X, y, locStr = removeHemisphere(X, y, lonlat, hemToRemove)
else:
    locStr = '_global'

# Take log of certain features
X = preprocessData(X)

# Adding random features to X
X = np.append(X, np.random.random([X.shape[0], nRandomFeatures]), axis=1)

# Remove correlated features as specified by adriana
X, featureLabels = extractUncorFeatures(X)

# Normalise Data
X, means, std = normaliseData(X)

# Map labels to groups
mapping = {'1': 1, '2': 1, '3': 1, '4': 2, '5': 3, '6': 4, '7': 5, '8': 5,
           '9': 9, '10': 10, '11': 11, '12': 12, '13': 3}
y_map = np.zeros(y.shape)
for i in range(y.shape[0]):
    y_map[i] = mapping[str(int(y[i]))]

# Remove prespecified labels
for remLab in removeLabels:
    X = X[y_map != remLab,:]
    y_map = y_map[y_map != remLab]

# Balance classes to be equally sized
if balanceDataFlag:
    X, y_map = utils.balanceData(X, y_map)

# Convert labels to 1's and 0's
y_map[y_map != lithoLabel] = 0
y_map[y_map == lithoLabel] = 1
nFeatures = X.shape[1]

# Saturate outliers
X = sat_outliers(X)

# Convert to original scale for plotting
Xorigscale = std*X + means

# Make featureID a base 0 indexed array
featureID = [feat-1 for feat in featureID]

# 2D and 3D plots
plotscatter2D(Xorigscale, y_map, featureID, featureLabels, locStr)
plotscatter3D(Xorigscale, y_map, featureID, featureLabels, locStr)

# Model data with a GP

plotprob2D(X,y_map,featureID,featureLabels,locStr, ntraining, Xorigscale)
# plotprobrevrand(X,y_map,featureID,featureLabels,locStr, ntraining, Xorigscale)
#

